from .common import *

__title__ = 'graphene_elastic.filter_backends.post_filter'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL-2.0-only OR LGPL-2.1-or-later'



from __future__ import unicode_literals

import inspect
from collections import OrderedDict

import elasticsearch_dsl
from elasticsearch_dsl import field as elasticsearch_fields

from graphene import Node
# from graphene.utils.trim_docstring import trim_docstring

__title__ = "graphene_elastic.converter"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    "get_document_fields",
    "get_field_description",
    "get_node_from_global_id",
    "get_type_for_document",
    "import_single_dispatch",
    "is_valid_elasticsearch_document",
)


def get_document_fields(document, excluding=None):
    excluding = excluding or []
    attributes = {"_id": elasticsearch_fields.Keyword()}
    for (
        attr_name,
        attr,
    ) in document._doc_type.mapping.properties.properties._d_.items():
        if attr_name in excluding:
            continue
        attributes[attr_name] = attr
    return OrderedDict(sorted(attributes.items()))


def is_valid_elasticsearch_document(document):
    return inspect.isclass(document) and (
        issubclass(document, elasticsearch_dsl.Document)
        or issubclass(document, elasticsearch_dsl.InnerDoc)
    )


def import_single_dispatch():
    try:
        from functools import singledispatch
    except ImportError:
        singledispatch = None

    if not singledispatch:
        try:
            from singledispatch import singledispatch
        except ImportError:
            pass

    if not singledispatch:
        raise Exception(
            "It seems your python version does not include "
            "functools.singledispatch. Please install the 'singledispatch' "
            "package. More information here: "
            "https://pypi.python.org/pypi/singledispatch"
        )

    return singledispatch


# noqa
def get_type_for_document(schema, document):
    types = schema.types.values()
    for _type in types:
        type_document = hasattr(_type, "_meta") and getattr(
            _type._meta, "document", None
        )
        if document == type_document:
            return _type


def get_field_description(field, registry=None):
    """
    Common metadata includes verbose_name and help_text.

    http://docs.mongoengine.org/apireference.html#fields
    """
    parts = []
    # if hasattr(field, 'document_type'):
    #     doc = trim_docstring(field.document_type.__doc__)
    #     if doc:
    #         parts.append(doc)
    # if hasattr(field, 'verbose_name'):
    #     parts.append(field.verbose_name.title())
    # if hasattr(field, 'help_text'):
    #     parts.append(field.help_text)

    # Adding string repr for doc-type. Name would stand for
    if hasattr(field, "name"):
        parts.append(field.name)

    # if field.db_field != field.name:
    #     name_format = "(%s)" if parts else "%s"
    #     parts.append(name_format % field.db_field)

    return "\n".join(parts)


def get_node_from_global_id(node, info, global_id):
    try:
        for interface in node._meta.interfaces:
            if issubclass(interface, Node):
                return interface.get_node_from_global_id(info, global_id)
    except AttributeError:
        return Node.get_node_from_global_id(info, global_id)



from copy import deepcopy
import enum
import graphene
from graphene_elastic.types.json_string import JSONString
from stringcase import pascalcase as to_pascal_case

from ...constants import DYNAMIC_CLASS_NAME_PREFIX
from ..base import BaseBackend

__title__ = 'graphene_elastic.filter_backends.ordering.common'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'HighlightFilterBackend',
)


def highlight_resolver(parent, args, context=None, info=None):
    """Highlight resolver.

    :param parent:
    :param args:
    :param context:
    :param info:
    :return:
    """
    # return parent.meta.highlight
    return parent.meta._d_.get('highlight')


class HighlightField(JSONString):
    """Highlight field."""


class HighlightFilterBackend(BaseBackend):
    """Highlight filter backend."""

    prefix = 'highlight'
    has_query_fields = True

    @property
    def highlight_fields(self):
        """Highlight filter fields."""
        highlight_fields = getattr(
            self.connection_field.type._meta.node._meta,
            'filter_backend_options',
            {}
        ).get('highlight_fields', {})
        return deepcopy(highlight_fields)

    def field_belongs_to(self, field_name):
        """Check if given filter field belongs to the backend.

        :param field_name:
        :return:
        """
        return field_name in self.highlight_fields

    def get_backend_query_fields(self,
                                 items,
                                 is_filterable_func,
                                 get_type_func):
        """Construct backend filtering fields.

        :param items:
        :param is_filterable_func:
        :param get_type_func:
        :return:
        """
        params = {}
        for field, value in items:
            if is_filterable_func(field):
                # Getting other backend specific fields (schema dependant)
                if self.field_belongs_to(field):
                    params.update({field: field})
        return {
            self.prefix: graphene.Argument(
                graphene.List(
                    graphene.Enum.from_enum(
                        enum.Enum(
                            "{}{}{}BackendEnum".format(
                                DYNAMIC_CLASS_NAME_PREFIX,
                                to_pascal_case(self.prefix),
                                self.connection_field.type.__name__
                            ),
                            params
                        )
                    )
                )
            )
        }

    def prepare_highlight_fields(self):
        """Prepare highlight fields.

        Possible structures:

            highlight_fields = {
                'title': {
                    'enabled': True,
                    'options': {
                        'pre_tags': ["<b>"],
                        'post_tags': ["</b>"],
                    }
                },
                'summary': {
                    'options': {
                        'fragment_size': 50,
                        'number_of_fragments': 3
                    }
                },
                'description': {},
            }

        Sample query would be:

            query {
              allPostDocuments(
                    search:{content:{value:"since"}, title:{value:"decide"}},
                    highlight:[category, content]
                ) {
                edges {
                  node {
                    title
                    content
                    highlight
                  }
                  cursor
                }
              }
            }

        :return: Filtering options.
        :rtype: dict
        """
        if not self.highlight_fields:
            return {}
        highlight_args = dict(self.args).get(self.prefix, [])

        highlight_fields = dict(self.highlight_fields)

        for field, options in self.highlight_fields.items():
            if 'enabled' in highlight_fields[field] \
                    or field not in highlight_args:
                highlight_fields[field]['enabled'] = True
            else:
                highlight_fields[field]['enabled'] = False

            if 'options' not in highlight_fields[field]:
                highlight_fields[field]['options'] = {}

        return highlight_fields

    def get_highlight_query_params(self):
        """Get highlight query params.

        :return: List of search query params.
        :rtype: list
        """
        highlight_args = dict(self.args).get(self.prefix, {})
        return highlight_args

    def filter(self, queryset):
        """Filter.

        :param queryset:
        :return:
        """
        highlight_query_params = self.get_highlight_query_params()
        highlight_fields = self.prepare_highlight_fields()

        for _field, _options in highlight_fields.items():
            if _field in highlight_query_params or _options['enabled']:
                queryset = queryset.highlight(_field, **_options['options'])

        return queryset

    def get_backend_document_fields(self):
        return {
            'highlight': graphene.Field(
                HighlightField,
                resolver=highlight_resolver
            ),
        }



# import uuid
from graphene import (
    # ID,
    Boolean,
    DateTime,
    # Dynamic,
    # Field,
    Float,
    Int,
    # List,
    # NonNull,
    String,
    # Union,
    # is_node,
)

from elasticsearch_dsl import (
    # InnerDoc,
    field as elasticsearch_fields,
)

from .utils import import_single_dispatch, get_field_description

__title__ = "graphene_elastic.converter"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"

singledispatch = import_single_dispatch()


class ElasticsearchConversionError(Exception):
    pass


@singledispatch
def convert_elasticsearch_field(field, registry=None):
    raise ElasticsearchConversionError(
        "Don't know how to convert the Elasticsearch field %s (%s)"
        % (field, field.__class__)
    )


@convert_elasticsearch_field.register(elasticsearch_fields.Text)
@convert_elasticsearch_field.register(elasticsearch_fields.Keyword)
def convert_field_to_string(field, registry=None):
    return String(
        description=get_field_description(field, registry),
        required=field._required,
    )


@convert_elasticsearch_field.register(elasticsearch_fields.Byte)
@convert_elasticsearch_field.register(elasticsearch_fields.Integer)
@convert_elasticsearch_field.register(elasticsearch_fields.Long)
@convert_elasticsearch_field.register(elasticsearch_fields.Short)
def convert_field_to_int(field, registry=None):
    return Int(
        description=get_field_description(field, registry),
        required=field._required,
    )


@convert_elasticsearch_field.register(elasticsearch_fields.Boolean)
def convert_field_to_boolean(field, registry=None):
    return Boolean(
        description=get_field_description(field, registry),
        required=field._required,
    )


@convert_elasticsearch_field.register(elasticsearch_fields.Double)
@convert_elasticsearch_field.register(elasticsearch_fields.HalfFloat)
@convert_elasticsearch_field.register(elasticsearch_fields.Float)
@convert_elasticsearch_field.register(elasticsearch_fields.ScaledFloat)
def convert_field_to_float(field, registry=None):
    return Float(
        description=get_field_description(field, registry),
        required=field._required,
    )


@convert_elasticsearch_field.register(elasticsearch_fields.Date)
def convert_field_to_datetime(field, registry=None):
    return DateTime(
        description=get_field_description(field, registry),
        required=field._required,
    )


# @convert_elasticsearch_field.register(elasticsearch_fields.Nested)
# def convert_nested_field_to_list(field, registry=None):
#     base_type = convert_field_to_jsonstring(field._doc_class)
#
#     if not isinstance(base_type, (List, NonNull)):
#         base_type = type(base_type)
#     return List(
#         base_type,
#         description=get_field_description(field, registry),
#         required=field._required
#     )


# @convert_elasticsearch_field.register(InnerDoc)
# def convert_field_to_inner_doc(field, registry=None):
#     from .types import JSONString
#
#     return JSONString(
#         description=get_field_description(field, registry),
#         required=field._required,
#     )


@convert_elasticsearch_field.register(elasticsearch_fields.Object)
@convert_elasticsearch_field.register(elasticsearch_fields.Nested)
def convert_field_to_jsonstring(field, registry=None):
    from .types import JSONString

    return JSONString(
        description=get_field_description(field, registry),
        required=field._required,
    )



import json
import logging
import sys

from .settings import graphene_settings

__title__ = 'graphene_elastic.logging'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'logger',
)

# Logger
logger = logging.getLogger(__name__)

# Get logging level. If not set, ERROR by default
try:
    logging_level = int(graphene_settings.LOGGING_LEVEL)
except ValueError as err:
    logging_level = logging.ERROR

# Set logging level for the logger
logger.setLevel(logging_level)

# This might be more flexible
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging_level)
formatter = logging.Formatter(
    '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
handler.setFormatter(formatter)
logger.addHandler(handler)


def debug_json(msg, *args, **kwargs):
    """
    Log 'msg % args' with severity 'DEBUG'.

    To pass exception information, use the keyword argument exc_info with
    a true value, e.g.

    logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
    """
    try:
        msg = json.dumps(msg)
    except Exception:
        pass
    return logger.debug(msg, *args, **kwargs)


logger.debug_json = debug_json



"""
Contains information about the current Elasticsearch version in use,
including (LTE and GTE).
"""

from distutils.version import LooseVersion

__title__ = 'graphene_elastic.versions'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = [
    'get_elasticsearch_version',
    'LOOSE_ELASTICSEARCH_VERSION',
    'LOOSE_ELASTICSEARCH_MINOR_VERSION',
]


def get_elasticsearch_version(default=(2, 0, 0)):
    """Get Elasticsearch version.

    :param default: Default value. Mainly added for building the docs
        when Elasticsearch is not running.
    :type default: tuple
    :return:
    :rtype: list
    """
    try:
        from elasticsearch_dsl import __version__
        return __version__
    except ImportError:
        return default


LOOSE_ELASTICSEARCH_VERSION = LooseVersion(
    '.'.join([str(__n) for __n in get_elasticsearch_version()])
)

LOOSE_ELASTICSEARCH_MINOR_VERSION = LooseVersion(
    '.'.join([str(i) for i in LOOSE_ELASTICSEARCH_VERSION.version[0:2]])
)

# Loose versions
LOOSE_VERSIONS = (
    '2.0',
    '2.1',
    '2.2',
    '5.0',
    '5.1',
    '5.2',
    '5.3',
    '5.4',
    '6.0',
    '6.1',
    '6.2',
    '6.3',
    '7.0',
    '7.1',
    '7.2',
    '7.3',
    '7.4',
    '8.0',
    '9.0',
)

for __v in LOOSE_VERSIONS:
    __var_name = 'LOOSE_VERSION_{0}'.format(__v.replace('.', '_'))
    globals()[__var_name] = LooseVersion(__v)
    __all__.append(__var_name)

# Exact versions
EXACT_VERSIONS = LOOSE_VERSIONS[:-1]

for __i, __v in enumerate(EXACT_VERSIONS):
    __l_cur = globals()['LOOSE_VERSION_{0}'
                        ''.format(LOOSE_VERSIONS[__i].replace('.', '_'))]
    __l_nxt = globals()['LOOSE_VERSION_{0}'
                        ''.format(LOOSE_VERSIONS[__i+1].replace('.', '_'))]
    __var_name = 'ELASTICSEARCH_{0}'.format(__v.replace('.', '_'))
    globals()[__var_name] = (__l_cur <= LOOSE_ELASTICSEARCH_VERSION < __l_nxt)
    __all__.append(__var_name)

# LTE list
LTE_VERSIONS = LOOSE_VERSIONS[:-1]

for __i, __v in enumerate(EXACT_VERSIONS):
    __l_cur = globals()['LOOSE_VERSION_{0}'
                        ''.format(LOOSE_VERSIONS[__i].replace('.', '_'))]
    __var_name = 'ELASTICSEARCH_LTE_{0}'.format(__v.replace('.', '_'))
    globals()[__var_name] = (LOOSE_ELASTICSEARCH_MINOR_VERSION <= __l_cur)
    __all__.append(__var_name)

# GTE list
GTE_VERSIONS = LOOSE_VERSIONS[:-1]

for __i, __v in enumerate(EXACT_VERSIONS):
    __l_cur = globals()['LOOSE_VERSION_{0}'
                        ''.format(LOOSE_VERSIONS[__i].replace('.', '_'))]
    __var_name = 'ELASTICSEARCH_GTE_{0}'.format(__v.replace('.', '_'))
    globals()[__var_name] = (
        LOOSE_ELASTICSEARCH_MINOR_VERSION >= __l_cur
    )
    __all__.append(__var_name)

__all__ = tuple(__all__)

# Clean up
try:
    del __l_cur
    del __l_nxt
    del __var_name
    del __i
    del __v
except NameError:
    pass

del LooseVersion



import unittest
import factories
from .base import BaseGrapheneElasticTestCase
from ..constants import ALL, VALUE

__all__ = (
    'SearchBackendElasticTestCase',
)


class SearchBackendElasticTestCase(BaseGrapheneElasticTestCase):

    def setUp(self):
        super(SearchBackendElasticTestCase, self).setUp()
        self.alice = "Alice"
        self.num_alice_posts = 9
        self.alice_posts = factories.PostFactory.create_batch(
            self.num_alice_posts
        )
        for _post in self.alice_posts:
            _post.content = "{} {} {}".format(
                self.faker.paragraph(),
                self.alice,
                self.faker.paragraph()
            )
            _post.save()

        self.beast = "Jabberwocky"
        self.num_beast_posts = 5
        self.beast_posts = factories.PostFactory.create_batch(
            self.num_beast_posts
        )
        for _post in self.beast_posts:
            _post.content = "{} {} {}".format(
                self.faker.paragraph(),
                self.beast,
                self.faker.paragraph()
            )
            _post.save()

        self.num_other_posts = 40
        self.other_posts = factories.PostFactory.create_batch(
            self.num_other_posts
        )
        for _post in self.other_posts:
            _post.save()

        self.sleep(2)

    def __test_search_content(self, search, num_posts):
        """Test search.

        content:{%s:"%s"}

        :param num_posts:
        :return:
        """
        query = """
        query {
          allPostDocuments(search:%s) {
            edges {
              node {
                category
                title
                comments
              }
            }
          }
        }
        """ % search
        print(query)
        executed = self.client.execute(query)
        self.assertEqual(
            len(executed['data']['allPostDocuments']['edges']),
            num_posts
        )

    def _test_search_content(self):
        """"Test search content.

        :return:
        """
        # Covering specific field lookups: `search:{title:{value:"Another"}}`
        with self.subTest('Test search the content on term "Django"'):
            self.__test_search_content(
                '{content:{%s:"%s"}}' % (VALUE, self.alice),
                self.num_alice_posts
            )
        with self.subTest('Test search the content on term "Elastic"'):
            self.__test_search_content(
                '{content:{%s:"%s"}}' % (VALUE, self.beast),
                self.num_beast_posts
            )

        # Covering all field lookups: `search:{query:"Another"}`
        with self.subTest('Test search the content on term "Django"'):
            self.__test_search_content(
                '{%s:"%s"}' % (ALL, self.alice),
                self.num_alice_posts
            )

        with self.subTest('Test search the content on term "Elastic"'):
            self.__test_search_content(
                '{%s:"%s"}' % (ALL, self.beast),
                self.num_beast_posts
            )

    def test_all(self):
        """Test all.

        Since we don't write in specific tests, it's more efficient to run
        them all from a single method in order to save on speed ups between
        tests.
        """
        self._test_search_content()


if __name__ == '__main__':
    unittest.main()



from copy import copy, deepcopy
import operator

import graphene
import six

from elasticsearch_dsl.query import Q
from stringcase import pascalcase as to_pascal_case

from ..base import BaseBackend
from ...constants import (
    DYNAMIC_CLASS_NAME_PREFIX,
    ALL,
    VALUE,
    BOOST,
)

__title__ = "graphene_elastic.filter_backends.search.common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("SearchFilterBackend",)


class SearchFilterBackend(BaseBackend):
    """Search filter backend."""

    prefix = "search"
    has_query_fields = True

    @property
    def search_fields(self):
        """Search filter fields."""
        search_fields = getattr(
            self.connection_field.type._meta.node._meta,
            'filter_backend_options',
            {}
        ).get('search_fields', {})
        return deepcopy(search_fields)

    @property
    def search_args_mapping(self):
        return {field: field for field, value in self.search_fields.items()}

    def field_belongs_to(self, field_name):
        """Check if given filter field belongs to the backend.

        :param field_name:
        :return:
        """
        return field_name in self.search_fields

    def get_backend_default_query_fields_params(self):
        """Get backend default filter params.

        :rtype: dict
        :return:
        """
        return {ALL: graphene.String()}

    def get_field_type(self, field_name, field_value, base_field_type):
        """Get field type.

        :return:
        """
        params = {
            VALUE: base_field_type,  # Value to search on. Required.
            BOOST: graphene.Int(),  # Boost the given field with. Optional.
        }
        return graphene.Argument(
            type(
                "{}{}{}{}".format(
                    DYNAMIC_CLASS_NAME_PREFIX,
                    to_pascal_case(self.prefix),
                    self.connection_field.type.__name__,
                    to_pascal_case(field_name)
                ),
                (graphene.InputObjectType,),
                params,
            )
        )

    def prepare_search_fields(self):
        """Prepare search fields.

        Possible structures:

            search_fields = {
                'title': {'boost': 4, 'field': 'title.raw'},
                'content': {'boost': 2},
                'category': None,
            }

        We shall finally have:

            search_fields = {
                'title': {
                    'field': 'title.raw',
                    'boost': 4
                },
                'content': {
                    'field': 'content',
                    'boost': 2
                },
                'category': {
                    'field': 'category'
                }
            }

        Sample query would be:

            {
              allPostDocuments(search:{query:"Another"}) {
                pageInfo {
                  startCursor
                  endCursor
                  hasNextPage
                  hasPreviousPage
                }
                edges {
                  cursor
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }


        :return: Filtering options.
        :rtype: dict
        """
        filter_args = dict(self.args).get(self.prefix)
        if not filter_args:
            return {}

        filter_fields = {}

        # {'query': '', 'title': {'query': '', 'boost': 1}}

        for field, _ in self.search_args_mapping.items():
            filter_fields.update({field: {}})
            options = self.search_fields.get(field)
            # For constructions like 'category': 'category.raw' we shall
            # have the following:
            #
            if options is None or isinstance(options, six.string_types):
                filter_fields.update(
                    {
                        field: {"field": options or field}
                    }
                )
            elif "field" not in options:
                filter_fields.update({field: options})
                filter_fields[field]["field"] = field
            else:
                filter_fields.update({field: options})

        return filter_fields

    def get_all_query_params(self):
        filter_args = dict(self.args).get(self.prefix)
        if not filter_args:
            return {}
        return filter_args

    def construct_search(self):
        """Construct search.

        We have to deal with two types of structures:

        Type 1:

        >>> search_fields = (
        >>>     'title',
        >>>     'description',
        >>>     'summary',
        >>> )

        Type 2:

        >>> search_fields = {
        >>>     'title': {'field': 'title', 'boost': 2},
        >>>     'description': None,
        >>>     'summary': None,
        >>> }

        In GraphQL shall be:

            query {
              allPostDocuments(search:{
                    query:"Another",
                    title:{value:"Another"}
                    summary:{value:"Another"}
                }) {
                pageInfo {
                  startCursor
                  endCursor
                  hasNextPage
                  hasPreviousPage
                }
                edges {
                  cursor
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        Or simply:

            query {
              allPostDocuments(search:{query:"education technology"}) {
                pageInfo {
                  startCursor
                  endCursor
                  hasNextPage
                  hasPreviousPage
                }
                edges {
                  cursor
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        all_query_params = self.get_all_query_params()
        search_fields = self.prepare_search_fields()
        _queries = []
        for search_field, value in all_query_params.items():
            if search_field == ALL:
                for field_name_param, field_name \
                        in self.search_args_mapping.items():
                    field_options = copy(search_fields[field_name])
                    field = field_options.pop("field", field_name)

                    if isinstance(value, dict):
                        # For constructions like:
                        # {'title': {'value': 'Produce', 'boost': 1}}
                        _query = value.pop(VALUE)
                        _field_options = copy(value)
                        value = _query
                        field_options.update(_field_options)
                    field_kwargs = {field: {"query": value}}
                    if field_options:
                        field_kwargs[field].update(field_options)
                    # The match query
                    _queries.append(Q("match", **field_kwargs))
            elif search_field in search_fields:
                field_options = copy(search_fields[search_field])
                field = field_options.pop("field", search_field)

                if isinstance(value, dict):
                    # For constructions like:
                    # {'title': {'value': 'Produce', 'boost': 1}}
                    _query = value.pop(VALUE)
                    _field_options = copy(value)
                    value = _query
                    field_options.update(_field_options)
                field_kwargs = {field: {"query": value}}
                if field_options:
                    field_kwargs[field].update(field_options)
                # The match query
                _queries.append(Q("match", **field_kwargs))

        return _queries

    # def construct_nested_search(self):
    #     """Construct nested search.
    #
    #     We have to deal with two types of structures:
    #
    #     Type 1:
    #
    #     >>> search_nested_fields = {
    #     >>>     'country': {
    #     >>>         'path': 'country',
    #     >>>         'fields': ['name'],
    #     >>>     },
    #     >>>     'city': {
    #     >>>         'path': 'country.city',
    #     >>>         'fields': ['name'],
    #     >>>     },
    #     >>> }
    #
    #     Type 2:
    #
    #     >>> search_nested_fields = {
    #     >>>     'country': {
    #     >>>         'path': 'country',
    #     >>>         'fields': [{'name': {'boost': 2}}]
    #     >>>     },
    #     >>>     'city': {
    #     >>>         'path': 'country.city',
    #     >>>         'fields': [{'name': {'boost': 2}}]
    #     >>>     },
    #     >>> }
    #
    #     :return: Updated queryset.
    #     :rtype: elasticsearch_dsl.search.Search
    #     """
    #     if (
    #         not hasattr(self, "search_nested_fields")
    #         or not self.search_nested_fields
    #     ):
    #         return []
    #
    #     # TODO: Support query boosting
    #
    #     query_params = self.prepare_search_fields()
    #     __queries = []
    #     for search_term in query_params:
    #         for label, options in self.search_nested_fields.items():
    #             queries = []
    #             path = options.get("path")
    #
    #             for _field in options.get("fields", []):
    #
    #                 # In case if we deal with structure 2
    #                 if isinstance(_field, dict):
    #                     # TODO: take options (such as boost) into
    #                     # consideration
    #                     field = "{}.{}".format(path, _field["name"])
    #                 # In case if we deal with structure 1
    #                 else:
    #                     field = "{}.{}".format(path, _field)
    #
    #                 field_kwargs = {field: search_term}
    #
    #                 queries.append(Q("match", **field_kwargs))
    #
    #             __queries.append(
    #                 Q(
    #                     "nested",
    #                     path=path,
    #                     query=six.moves.reduce(operator.or_, queries),
    #                 )
    #             )
    #
    #     return __queries

    def filter(self, queryset):
        """Filter.

        :param queryset:
        :return:
        """
        _queries = []

        _search = self.construct_search()
        if _search:
            _queries.extend(_search)

        # _nested_search = self.construct_nested_search()
        # if _nested_search:
        #     _queries.extend(_nested_search)

        if _queries:
            queryset = queryset.query("bool", should=_queries)
        return queryset



from typing import Callable
import operator
from elasticsearch_dsl.query import Q
import six

from ...constants import (
    FALSE_VALUES,
    TRUE_VALUES,
    LOWER,
    UPPER,
    GTE,
    LTE,
    BOOST,
)

__title__ = "graphene_elastic.filter_backends.filtering.mixins"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("FilteringFilterMixin",)


class FilteringFilterMixin(object):
    """Filtering filter mixin."""

    apply_query: Callable
    apply_filter: Callable
    split_lookup_complex_value: Callable

    @classmethod
    def get_range_param_value(cls, value):
        """Get range param value.

        :param value:
        :type value:
            graphene_elastic.filter_backends.filtering.queries.InputObjectType
        :return:
        """
        if not value:
            return None

        return (
            value.decimal
            or value.float
            or value.int
            or value.date
            or value.datetime
        )

    @classmethod
    def get_range_params(cls, value, options):
        """Get params for `range` query.

        Syntax:

            TODO

        Example:

            {
              allPostDocuments(filter:{numViews:{range:{
                    lower:{decimal:"100"},
                    upper:{decimal: "200"},
                    boost:"2.0"
                  }}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param value:
        :param options:
        :type value:
            graphene_elastic.filter_backends.filtering.queries.InputObjectType
        :type options: dict
        :return: Params to be used in `range` query.
        :rtype: dict
        """
        if LOWER not in value:
            return {}

        params = {GTE: cls.get_range_param_value(value.get(LOWER, None))}

        if UPPER in value:
            params.update({
                LTE: cls.get_range_param_value(value.get(UPPER, None))
            })

        if BOOST in value:
            params.update({BOOST: float(value.get(BOOST, None))})

        return params

    @classmethod
    def get_gte_lte_params(cls, value, lookup, options):
        """Get params for `gte`, `gt`, `lte` and `lt` query.

        Syntax:

            TODO

        Example:

            {
              allPostDocuments(filter:{numViews:{
                    gt:{decimal:"100"},
                    lt:{decimal:"200"}
                }}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param value:
        :param lookup:
        :param options:
        :type value:
            graphene_elastic.filter_backends.filtering.queries.InputObjectType
        :type lookup: str
        :type options: dict
        :return: Params to be used in `range` query.
        :rtype: dict
        """
        if not value:
            return {}

        _value = cls.get_range_param_value(value)

        if not _value:
            return {}

        params = {lookup: _value}

        if BOOST in options:
            params.update({BOOST: float(options.get(BOOST, None))})

        return params

    @classmethod
    def apply_filter_term(cls, queryset, options, value):
        """Apply `term` filter.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{category:{term:"Python"}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                    comments
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["term"],
            kwargs={options["field"]: value},
        )

    @classmethod
    def apply_filter_terms(cls, queryset, options, value):
        """Apply `terms` filter.

        Syntax:

            TODO

        Note, that number of values is not limited.

        Example:

        query {
          allPostDocuments(filter:{category:{
                terms:["Python", "Django"]
              }}) {
            edges {
              node {
                category
                title
                content
                numViews
                comments
              }
            }
          }
        }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: mixed: either str or iterable (list, tuple).
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        # If value is a list or a tuple, we use it as is.
        if isinstance(value, (list, tuple)):
            __values = value

        # Otherwise, we consider it to be a string and split it further.
        else:
            __values = cls.split_lookup_complex_value(value)

        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["terms"],
            kwargs={options["field"]: __values},
        )

    @classmethod
    def apply_filter_range(cls, queryset, options, value):
        """Apply `range` filter.

         Syntax:

            TODO

        Example:

            {
              allPostDocuments(filter:{numViews:{range:{
                    lower:{decimal:"100"},
                    upper:{decimal:"200"}
                  }}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }


        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={
                options["field"]: cls.get_range_params(
                    value,
                    options.get('options', {})
                )
            },
        )

    @classmethod
    def apply_query_exists(cls, queryset, options, value):
        """Apply `exists` filter.

        Syntax:

            TODO

        Example:

            {
              allPostDocuments(filter:{category:{exists:true}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        _value_lower = value  # TODO: clean up?
        if _value_lower in TRUE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[Q("exists", field=options["field"])],
            )
        elif _value_lower in FALSE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[~Q("exists", field=options["field"])],
            )
        return queryset

    @classmethod
    def apply_filter_prefix(cls, queryset, options, value):
        """Apply `prefix` filter.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{category:{prefix:"Pyth"}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                    comments
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["prefix"],
            kwargs={options["field"]: value},
        )

    @classmethod
    def apply_query_wildcard(cls, queryset, options, value):
        """Apply `wildcard` filter.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{category:{wildcard:"*ytho*"}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                    comments
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_query(
            queryset=queryset,
            options=options,
            args=[Q("wildcard", **{options["field"]: value})],
        )

    @classmethod
    def apply_query_contains(cls, queryset, options, value):
        """Apply `contains` filter.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{category:{contains:"tho"}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }


        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_query(
            queryset=queryset,
            options=options,
            args=[Q("wildcard", **{options["field"]: "*{}*".format(value)})],
        )

    @classmethod
    def apply_query_endswith(cls, queryset, options, value):
        """Apply `endswith` filter.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{category:{endsWith:"thon"}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_query(
            queryset=queryset,
            options=options,
            args=[Q("wildcard", **{options["field"]: "*{}".format(value)})],
        )

    @classmethod
    def apply_query_in(cls, queryset, options, value):
        """Apply `in` functional query.

        Syntax:

            TODO

        Note, that number of values is not limited.

        Example:

            query {
              allPostDocuments(postFilter:{tags:{
                    in:["photography", "models"]
                }}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                    tags
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        # If value is a list or a tuple, we use it as is.
        if isinstance(value, (list, tuple)):
            _values = value

        # Otherwise, we consider it to be a string and split it further.
        else:
            _values = cls.split_lookup_complex_value(value)

        _queries = []
        for _value in _values:
            _queries.append(Q("term", **{options["field"]: _value}))

        if _queries:
            queryset = cls.apply_query(
                queryset=queryset,
                options=options,
                args=[six.moves.reduce(operator.or_, _queries)],
            )

        return queryset

    @classmethod
    def apply_query_gt(cls, queryset, options, value):
        """Apply `gt` functional query.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{numViews:{
                    gt:{decimal:"100"}
                }}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={
                options["field"]: cls.get_gte_lte_params(
                    value,
                    "gt",
                    options.get('options', {})
                )
            },
        )

    @classmethod
    def apply_query_gte(cls, queryset, options, value):
        """Apply `gte` functional query.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{numViews:{
                    gte:{decimal:"100"}
                }}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={
                options["field"]: cls.get_gte_lte_params(
                    value,
                    "gte",
                    options.get('options', {})
                )
            },
        )

    @classmethod
    def apply_query_lt(cls, queryset, options, value):
        """Apply `lt` functional query.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{numViews:{
                    lt:{decimal:"200"}
                }}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={
                options["field"]: cls.get_gte_lte_params(
                    value,
                    "lt",
                    options.get('options', {})
                )
            },
        )

    @classmethod
    def apply_query_lte(cls, queryset, options, value):
        """Apply `lte` functional query.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{numViews:{
                    lte:{decimal:"200"}
                }}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return cls.apply_filter(
            queryset=queryset,
            options=options,
            args=["range"],
            kwargs={
                options["field"]: cls.get_gte_lte_params(
                    value,
                    "lte",
                    options.get('options', {})
                )
            },
        )

    @classmethod
    def apply_query_isnull(cls, queryset, options, value):
        """Apply `isnull` functional query.

        Syntax:

            TODO

        Example:

            query {
              allPostDocuments(filter:{category:{isNull:true}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                    comments
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        _value_lower = value  # TODO: clean up?
        if _value_lower in TRUE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[~Q("exists", field=options["field"])],
            )
        elif _value_lower in FALSE_VALUES:
            return cls.apply_query(
                queryset=queryset,
                options=options,
                args=[Q("exists", field=options["field"])],
            )
        return queryset

    @classmethod
    def apply_query_exclude(cls, queryset, options, value):
        """Apply `exclude` functional query.

        Syntax:

            TODO

        Note, that number of values is not limited.

        Example:

            query {
              allPostDocuments(filter:{category:{exclude:"Python"}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        Or exclude multiple terms at once:

            query {
              allPostDocuments(filter:{category:{exclude:["Ruby", "Java"]}}) {
                edges {
                  node {
                    category
                    title
                    content
                    numViews
                  }
                }
              }
            }

        :param queryset: Original queryset.
        :param options: Filter options.
        :param value: value to filter on.
        :type queryset: elasticsearch_dsl.search.Search
        :type options: dict
        :type value: str
        :return: Modified queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        if not isinstance(value, (list, tuple)):
            __values = cls.split_lookup_complex_value(value)
        else:
            __values = value

        __queries = []
        for __value in __values:
            __queries.append(~Q("term", **{options["field"]: __value}))

        if __queries:
            queryset = cls.apply_query(
                queryset=queryset,
                options=options,
                args=[six.moves.reduce(operator.and_, __queries)],
            )

        return queryset



"""
Some overrides of the original ``graphql_relay.connection.connection``
module. For sanity and ease of updates/sync with modifications from upstream,
this module isn't formatted in accordance with the rest of the package.
Pull requests code-style changes wouldn't be accepted.
"""
import re
from collections import (
    # Iterable,
    OrderedDict,
)
# from functools import partial

# from graphql_relay import connection_from_list

from graphene.types import (
    # Boolean,
    Enum,
    # Int,
    Interface,
    List,
    NonNull,
    Scalar,
    String,
    Union,
)
from graphene.types.field import Field
from graphene.types.objecttype import ObjectType, ObjectTypeOptions
# from graphene.utils.thenables import maybe_thenable
# from graphene.relay.node import is_node

from graphene.relay import PageInfo


__title__ = 'graphene_elastic.relay.connection'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'Connection',
    'ConnectionOptions',
)


# class PageInfo(ObjectType):
#     class Meta:
#         description = (
#             "The Relay compliant `PageInfo` type, containing data necessary to"
#             " paginate this connection."
#         )
#
#     has_next_page = Boolean(
#         required=True,
#         name="hasNextPage",
#         description="When paginating forwards, are there more items?",
#     )
#
#     has_previous_page = Boolean(
#         required=True,
#         name="hasPreviousPage",
#         description="When paginating backwards, are there more items?",
#     )
#
#     start_cursor = String(
#         name="startCursor",
#         description="When paginating backwards, the cursor to continue.",
#     )
#
#     end_cursor = String(
#         name="endCursor",
#         description="When paginating forwards, the cursor to continue.",
#     )


class Facets(ObjectType):
    class Meta:
        description = (
            "The Relay compliant `Facets` type, containing facets data."
        )


class ConnectionOptions(ObjectTypeOptions):
    node = None


class Connection(ObjectType):
    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(cls, node=None, name=None, **options):
        _meta = ConnectionOptions(cls)
        assert node, "You have to provide a node in {}.Meta".format(cls.__name__)
        assert isinstance(node, NonNull) or issubclass(
            node, (Scalar, Enum, ObjectType, Interface, Union, NonNull)
        ), ('Received incompatible node "{}" for Connection {}.').format(
            node, cls.__name__
        )

        base_name = re.sub("Connection$", "", name or cls.__name__) or node._meta.name
        if not name:
            name = "{}Connection".format(base_name)

        edge_class = getattr(cls, "Edge", None)
        _node = node

        class EdgeBase(object):
            node = Field(_node, description="The item at the end of the edge")
            cursor = String(required=True, description="A cursor for use in pagination")

        class EdgeMeta:
            description = "A Relay edge containing a `{}` and its cursor.".format(
                base_name
            )

        edge_name = "{}Edge".format(base_name)
        if edge_class:
            edge_bases = (edge_class, EdgeBase, ObjectType)
        else:
            edge_bases = (EdgeBase, ObjectType)

        edge = type(edge_name, edge_bases, {"Meta": EdgeMeta})
        cls.Edge = edge

        from graphene_elastic.types.json_string import JSONString

        options["name"] = name
        _meta.node = node
        _meta.fields = OrderedDict(
            [
                (
                    "page_info",
                    Field(
                        PageInfo,
                        name="pageInfo",
                        required=True,
                        description="Pagination data for this connection.",
                    ),
                ),
                (
                    "edges",
                    Field(
                        NonNull(List(edge)),
                        description="Contains the nodes in this connection.",
                    ),
                ),
                # TODO: Construct this dynamically from the filter backends.
                (
                    "facets",
                    Field(
                        JSONString,
                        name="facets",
                        required=False,
                        description="Pagination data for this connection.",
                    ),
                ),
            ]
        )
        # for backend_cls in backends:
        #     if backend_cls.has_connection_fields:
        #         backend = backend_cls()
        #         connection_fields_type = backend.get_backend_connection_fields_type()
        #         if connection_fields_type:
        #             try:
        #                 _meta.fields.update(connection_fields_type)
        #             except:
        #                 pass
        return super(Connection, cls).__init_subclass_with_meta__(
            _meta=_meta, **options
        )


# class IterableConnectionField(Field):
#     def __init__(self, type, *args, **kwargs):
#         kwargs.setdefault("before", String())
#         kwargs.setdefault("after", String())
#         kwargs.setdefault("first", Int())
#         kwargs.setdefault("last", Int())
#         super(IterableConnectionField, self).__init__(type, *args, **kwargs)
#
#     @property
#     def type(self):
#         type = super(IterableConnectionField, self).type
#         connection_type = type
#         if isinstance(type, NonNull):
#             connection_type = type.of_type
#
#         if is_node(connection_type):
#             raise Exception(
#                 "ConnectionFields now need a explicit ConnectionType for Nodes.\n"
#                 "Read more: https://github.com/graphql-python/graphene/blob/v2.0.0/UPGRADE-v2.0.md#node-connections"
#             )
#
#         assert issubclass(connection_type, Connection), (
#             '{} type have to be a subclass of Connection. Received "{}".'
#         ).format(self.__class__.__name__, connection_type)
#         return type
#
#     @classmethod
#     def resolve_connection(cls, connection_type, args, resolved):
#         if isinstance(resolved, connection_type):
#             return resolved
#
#         assert isinstance(resolved, Iterable), (
#             "Resolved value from the connection field have to be iterable or instance of {}. "
#             'Received "{}"'
#         ).format(connection_type, resolved)
#         connection = connection_from_list(
#             resolved,
#             args,
#             connection_type=connection_type,
#             edge_type=connection_type.Edge,
#             pageinfo_type=PageInfo,
#         )
#         connection.iterable = resolved
#         return connection
#
#     @classmethod
#     def connection_resolver(cls, resolver, connection_type, root, info, **args):
#         resolved = resolver(root, info, **args)
#
#         if isinstance(connection_type, NonNull):
#             connection_type = connection_type.of_type
#
#         on_resolve = partial(cls.resolve_connection, connection_type, args)
#         return maybe_thenable(resolved, on_resolve)
#
#     def get_resolver(self, parent_resolver):
#         resolver = super(IterableConnectionField, self).get_resolver(parent_resolver)
#         return partial(self.connection_resolver, resolver, self.type)
#
#
# ConnectionField = IterableConnectionField



from __future__ import unicode_literals
import importlib
import json
import os
import logging

import six

__title__ = "graphene_elastic.settings"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("graphene_settings",)


DEFAULTS = {
    "SCHEMA": None,
    "SCHEMA_OUTPUT": "schema.json",
    "SCHEMA_INDENT": 2,
    # "MIDDLEWARE": (),
    # Set to True if the connection fields must have
    # either the first or last argument
    "RELAY_CONNECTION_ENFORCE_FIRST_OR_LAST": False,
    # Max items returned in ConnectionFields / FilterConnectionFields
    "RELAY_CONNECTION_MAX_LIMIT": 100,
    "LOGGING_LEVEL": logging.ERROR,
}

# List of settings that may be in string import notation.
IMPORT_STRINGS = ("MIDDLEWARE", "SCHEMA")


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, six.string_types):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split(".")
        module_path, class_name = ".".join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '%s' for Graphene setting '%s'. %s: %s." % (
            val,
            setting_name,
            e.__class__.__name__,
            e,
        )
        raise ImportError(msg)


def set_setting(key, value):
    """Set setting.

    :param key:
    :param value:
    :return:
    """
    if not isinstance(value, six.string_types):
        json_value = json.dumps(value)
        os.environ.setdefault(key, json_value)
    else:
        os.environ.setdefault(key, value)


def get_setting(key, default=None):
    """Get setting.

    :param key:
    :param default:
    :return:
    """
    value = os.environ.get(key, default=default)
    try:
        json_value = json.loads(value)
        return json_value
    except Exception as err:
        print(err)
    return value


class GrapheneSettings(object):
    """
    A settings object, that allows API settings to be accessed as properties.
    For example:
        from graphene_elastic.settings import settings
        print(settings.SCHEMA)
    Any setting with string import paths will be automatically resolved
    and return the class, rather than the string literal.
    """

    def __init__(self, user_settings=None, defaults=None, import_strings=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = get_setting("GRAPHENE_ELASTIC", {})

        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Graphene setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        setattr(self, attr, val)
        return val


graphene_settings = GrapheneSettings(None, DEFAULTS, IMPORT_STRINGS)



"""
Some overrides of the original ``graphql_relay.connection.connectiontypes``
module. For sanity and ease of updates/sync with modifications from upstream,
this module isn't formatted in accordance with the rest of the package.
Pull requests code-style changes wouldn't be accepted.
"""

__title__ = 'graphene_elastic.relay.connectiontypes'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'Connection',
    'Facets',
)


class Connection(object):

    def __init__(self, edges, page_info, facets=None):
        self.edges = edges
        self.page_info = page_info
        self.facets = facets

    def to_dict(self):
        return {
            'edges': [e.to_dict() for e in self.edges],
            'pageInfo': self.page_info.to_dict(),
            'facets': {}
        }


class Facets(object):

    def __init__(self, facets):
        self.facets = facets

    def to_dict(self):
        return {
            'facets': self.facets,
        }


# class PageInfo(object):
#
#     def __init__(self, start_cursor="", end_cursor="",
#                  has_previous_page=False, has_next_page=False):
#         self.startCursor = start_cursor
#         self.endCursor = end_cursor
#         self.hasPreviousPage = has_previous_page
#         self.hasNextPage = has_next_page
#
#     def to_dict(self):
#         return {
#             'startCursor': self.startCursor,
#             'endCursor': self.endCursor,
#             'hasPreviousPage': self.hasPreviousPage,
#             'hasNextPage': self.hasNextPage,
#         }
#
#
# class Edge(object):
#
#     def __init__(self, node, cursor):
#         self.node = node
#         self.cursor = cursor
#
#     def to_dict(self):
#         return {
#             'node': self.node,
#             'cursor': self.cursor,
#         }



from copy import deepcopy
import enum
import graphene
from stringcase import pascalcase as to_pascal_case

from ..base import BaseBackend
from ...constants import DYNAMIC_CLASS_NAME_PREFIX

__title__ = 'graphene_elastic.filter_backends.ordering.common'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'SourceFilterBackend',
)


class SourceFilterBackend(BaseBackend):
    """Source filter backend."""

    prefix = 'source'
    has_query_fields = True

    @property
    def source_fields(self):
        """Source filter fields."""
        source_fields = getattr(
            self.connection_field.type._meta.node._meta,
            'filter_backend_options',
            {}
        ).get('source_fields', {})
        return deepcopy(source_fields)

    def get_backend_query_fields(self,
                                 items,
                                 is_filterable_func,
                                 get_type_func):
        """Construct backend filtering fields.

        :param items:
        :param is_filterable_func:
        :param get_type_func:
        :return:
        """
        # Note, that this is not the same as ``self.source_fields``.
        _keys = list(
            self.connection_field.type._meta.node._meta.fields.keys()
        )
        _keys.remove('_id')
        params = zip(_keys, _keys)
        return {
            self.prefix: graphene.Argument(
                graphene.List(
                    graphene.Enum.from_enum(
                        enum.Enum(
                            "{}{}{}BackendEnum".format(
                                DYNAMIC_CLASS_NAME_PREFIX,
                                to_pascal_case(self.prefix),
                                self.connection_field.type.__name__
                            ),
                            params
                        )
                    )
                )
            )
        }

    def prepare_source_fields(self):
        """Prepare source fields.

        Possible structures:

            source_fields = ["title"]

        Or:

            search_fields = ["title", "author.*"]

        Or:

            source = {
                "includes": ["title", "author.*"],
                "excludes": [ "*.description" ]
            }

        :return: Filtering options.
        :rtype: dict
        """
        source_args = dict(self.args).get(self.prefix, [])

        source_fields = dict(self.source_fields)

        if source_args:
            return source_args
        return source_fields

    def filter(self, queryset):
        """Filter.

        :param queryset:
        :return:
        """
        source_fields = self.prepare_source_fields()

        if source_fields:
            queryset = queryset.source(source_fields)

        return queryset



from copy import deepcopy
import graphene
import six
from stringcase import pascalcase as to_pascal_case

from ..base import BaseBackend
from ...constants import (
    ALL_LOOKUP_FILTERS_AND_QUERIES,
    DYNAMIC_CLASS_NAME_PREFIX,
    LOOKUP_FILTER_EXISTS,
    LOOKUP_FILTER_PREFIX,
    LOOKUP_FILTER_RANGE,
    LOOKUP_FILTER_TERM,
    LOOKUP_FILTER_TERMS,
    LOOKUP_FILTER_WILDCARD,
    LOOKUP_QUERY_CONTAINS,
    LOOKUP_QUERY_ENDSWITH,
    LOOKUP_QUERY_EXCLUDE,
    LOOKUP_QUERY_GT,
    LOOKUP_QUERY_GTE,
    LOOKUP_QUERY_IN,
    LOOKUP_QUERY_ISNULL,
    LOOKUP_QUERY_LT,
    LOOKUP_QUERY_LTE,
    LOOKUP_QUERY_STARTSWITH,
    VALUE,
)
from .mixins import FilteringFilterMixin
from .queries import LOOKUP_FILTER_MAPPING

__title__ = "graphene_elastic.filter_backends.filtering.common"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("FilteringFilterBackend",)


class FilteringFilterBackend(BaseBackend, FilteringFilterMixin):
    """Filtering filter backend."""

    prefix = "filter"
    has_query_fields = True

    @property
    def filter_fields(self):
        """Filtering filter fields."""
        filter_fields = getattr(
            self.connection_field.type._meta.node._meta,
            'filter_backend_options',
            {}
        ).get('filter_fields', {})

        return deepcopy(filter_fields)

    @property
    def filter_args_mapping(self):
        return {field: field for field, value in self.filter_fields.items()}

    def field_belongs_to(self, field_name):
        """Check if given filter field belongs to the backend.

        :param field_name:
        :return:
        """
        return field_name in self.filter_fields

    def get_backend_query_fields(self,
                                 items,
                                 is_filterable_func,
                                 get_type_func):
        """Fail proof override.

        :param items:
        :param is_filterable_func:
        :param get_type_func:
        :return:
        """
        if not self.filter_fields:
            return {}

        return super(FilteringFilterBackend, self).get_backend_query_fields(
            items=items,
            is_filterable_func=is_filterable_func,
            get_type_func=get_type_func
        )

    def get_field_type(self, field_name, field_value, base_field_type):
        """Get field type.

        :return:
        """
        if not self.filter_fields:
            return None

        field_options = self.get_field_options(field_name)
        if isinstance(field_options, dict) and "lookups" in field_options:
            lookups = field_options.get("lookups", [])
        else:
            lookups = list(ALL_LOOKUP_FILTERS_AND_QUERIES)

        params = {VALUE: base_field_type}
        for lookup in lookups:
            query_cls = LOOKUP_FILTER_MAPPING.get(lookup)
            if not query_cls:
                continue
            params.update({lookup: query_cls()})

        return graphene.Argument(
            type(
                "{}{}{}{}".format(
                    DYNAMIC_CLASS_NAME_PREFIX,
                    to_pascal_case(self.prefix),
                    self.connection_field.type.__name__,
                    to_pascal_case(field_name)
                ),
                (graphene.InputObjectType,),
                params,
            )
        )

    # def get_field_options(self, field_name):
    #     """Get field options."""
    #     if field_name in self.filter_fields:
    #         return self.filter_fields[field_name]
    #     return {}

    def prepare_filter_fields(self):
        """Prepare filter fields.

        Possible structures:

            filter_fields = {
                'title': {
                    'field': 'title.raw',
                    'lookups': [
                        LOOKUP_FILTER_TERM,
                        LOOKUP_FILTER_TERMS,
                        LOOKUP_FILTER_PREFIX,
                        LOOKUP_FILTER_WILDCARD,
                        LOOKUP_QUERY_IN,
                        LOOKUP_QUERY_EXCLUDE,
                    ],
                    'default_lookup': LOOKUP_FILTER_TERM,
                },
                'category': 'category.raw',
            }

        We shall finally have:

            filter_fields = {
                'title': {
                    'field': 'title.raw',
                    'lookups': [
                        LOOKUP_FILTER_TERM,
                        LOOKUP_FILTER_TERMS,
                        LOOKUP_FILTER_PREFIX,
                        LOOKUP_FILTER_WILDCARD,
                        LOOKUP_QUERY_IN,
                        LOOKUP_QUERY_EXCLUDE,
                    ],
                    'default_lookup': LOOKUP_FILTER_TERM,
                },
                'category': {
                    'field': 'category.raw',
                    'lookups': [
                        LOOKUP_FILTER_TERM,
                        LOOKUP_FILTER_TERMS,
                        LOOKUP_FILTER_PREFIX,
                        LOOKUP_FILTER_WILDCARD,
                        LOOKUP_QUERY_IN,
                        LOOKUP_QUERY_EXCLUDE,
                        ...
                        # All other lookups
                    ],
                    'default_lookup': LOOKUP_FILTER_TERM,
                }
            }
        """
        filter_args = dict(self.args).get(self.prefix)
        if not filter_args:
            return {}

        filter_fields = {}

        for arg, value in filter_args.items():
            field = self.filter_args_mapping.get(arg, None)
            if field is None:
                continue
            filter_fields.update({field: {}})
            options = self.filter_fields.get(field)
            # For constructions like 'category': 'category.raw' we shall
            # have the following:
            # TODO: Make sure to use custom (user specified) lookups
            if options is None or isinstance(options, six.string_types):
                filter_fields.update(
                    {
                        field: {
                            "field": options or field,
                            "default_lookup": LOOKUP_FILTER_TERM,
                            "lookups": tuple(ALL_LOOKUP_FILTERS_AND_QUERIES),
                        }
                    }
                )
            elif "field" not in options:
                filter_fields.update({field: options})
                filter_fields[field]["field"] = field
            else:
                filter_fields.update({field: options})

            if (
                field in filter_fields
                and "lookups" not in filter_fields[field]
            ):
                filter_fields[field].update(
                    {"lookups": tuple(ALL_LOOKUP_FILTERS_AND_QUERIES)}
                )
        return filter_fields

    def prepare_query_params(self):
        """Prepare query params.

        :return:
        """
        filter_args = dict(self.args).get(self.prefix)
        if not filter_args:
            return {}

        query_params = {}

        for arg, filters in filter_args.items():
            field = self.filter_args_mapping.get(arg, None)
            if field is None:
                continue
            query_params[field] = filters
        return query_params

    def get_field_lookup_param(self, field_name):
        """Get field lookup param.

        :param field_name:
        :return:
        """
        field_options = dict(self.args) \
            .get(self.prefix, {}) \
            .get(field_name, {})
        return field_options.get("lookup", None)

    def get_field_options(self, field_name):
        """Get field option params.

        :param field_name:
        :return:
        """
        return dict(self.args).get(self.prefix, {}).get(field_name, {})

    def get_filter_query_params(self):
        """Get query params to be filtered on.

        We can either specify it like this:

            query_params = {
                'category': {
                    'value': 'Elastic',
                }
            }

        Or using specific lookup:

            query_params = {
                'category': {
                    'term': 'Elastic',
                    'range': {
                        'lower': Decimal('3.0')
                    }
                }
            }

        Note, that `value` would only work on simple types (string, integer,
        decimal). For complex types you would have to use complex param
        anyway. Therefore, it should be forbidden to set `default_lookup` to a
        complex field type.

        Sample values:

            query_params = {
                'category': {
                    'value': 'Elastic',
                }
            }

            filter_fields = {
                'category': {
                    'field': 'category.raw',
                    'default_lookup': 'term',
                    'lookups': (
                        'term',
                        'terms',
                        'range',
                        'exists',
                        'prefix',
                        'wildcard',
                        'contains',
                        'in',
                        'gt',
                        'gte',
                        'lt',
                        'lte',
                        'starts_with',
                        'ends_with',
                        'is_null',
                        'exclude'
                    )
                }
            }

            field_name = 'category'
        """
        query_params = self.prepare_query_params()  # Shall be fixed

        filter_query_params = {}
        filter_fields = self.prepare_filter_fields()  # Correct

        for field_name, lookup_params in query_params.items():

            if field_name in filter_fields:
                filter_query_params[field_name] = []
                valid_lookups = filter_fields[field_name]["lookups"]

                # If we have default lookup given use it as a default and
                # do not require further suffix specification.
                default_lookup = None
                if "default_lookup" in filter_fields[field_name]:
                    default_lookup = \
                        filter_fields[field_name]["default_lookup"]

                for lookup_param, lookup_options in lookup_params.items():
                    lookup = None
                    if lookup_param == VALUE:
                        if default_lookup is not None:
                            lookup = default_lookup
                    elif lookup_param in valid_lookups:
                        lookup = lookup_param

                    if lookup_options is not None:
                        filter_query_params[field_name].append({
                            "lookup": lookup,
                            "values": lookup_options,
                            "field": filter_fields[field_name].get(
                                "field", field_name
                            ),
                            "type": self.doc_type.mapping.properties.name,
                        })
        return filter_query_params

    def filter(self, queryset):
        """Filter."""
        filter_query_params = self.get_filter_query_params()

        for options in filter_query_params.values():

            # For all other cases, when we don't have multiple values,
            # we follow the normal flow.
            for option in options:

                if option["lookup"] == LOOKUP_FILTER_TERMS:
                    queryset = self.apply_filter_terms(
                        queryset,
                        option,
                        option['values']
                    )

                # `prefix` filter lookup
                elif option["lookup"] in (
                    LOOKUP_FILTER_PREFIX,
                    LOOKUP_QUERY_STARTSWITH,
                ):
                    queryset = self.apply_filter_prefix(
                        queryset,
                        option,
                        option['values']
                    )

                # `range` filter lookup
                elif option["lookup"] == LOOKUP_FILTER_RANGE:
                    queryset = self.apply_filter_range(
                        queryset,
                        option,
                        option['values']
                    )

                # `exists` filter lookup
                elif option["lookup"] == LOOKUP_FILTER_EXISTS:
                    queryset = self.apply_query_exists(
                        queryset,
                        option,
                        option['values']
                    )

                # `wildcard` filter lookup
                elif option["lookup"] == LOOKUP_FILTER_WILDCARD:
                    queryset = self.apply_query_wildcard(
                        queryset,
                        option,
                        option['values']
                    )

                # `contains` filter lookup
                elif option["lookup"] == LOOKUP_QUERY_CONTAINS:
                    queryset = self.apply_query_contains(
                        queryset,
                        option,
                        option['values']
                    )

                # `in` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_IN:
                    queryset = self.apply_query_in(
                        queryset,
                        option,
                        option['values']
                    )

                # `gt` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_GT:
                    queryset = self.apply_query_gt(
                        queryset,
                        option,
                        option['values']
                    )

                # `gte` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_GTE:
                    queryset = self.apply_query_gte(
                        queryset,
                        option,
                        option['values']
                    )

                # `lt` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_LT:
                    queryset = self.apply_query_lt(
                        queryset,
                        option,
                        option['values']
                    )

                # `lte` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_LTE:
                    queryset = self.apply_query_lte(
                        queryset,
                        option,
                        option['values']
                    )

                # `endswith` filter lookup
                elif option["lookup"] == LOOKUP_QUERY_ENDSWITH:
                    queryset = self.apply_query_endswith(
                        queryset,
                        option,
                        option['values']
                    )

                # `isnull` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_ISNULL:
                    queryset = self.apply_query_isnull(
                        queryset,
                        option,
                        option['values']
                    )

                # `exclude` functional query lookup
                elif option["lookup"] == LOOKUP_QUERY_EXCLUDE:
                    queryset = self.apply_query_exclude(
                        queryset,
                        option,
                        option['values']
                    )

                # `term` filter lookup. This is default if no `default_lookup`
                # option has been given or explicit lookup provided.
                else:
                    queryset = self.apply_filter_term(
                        queryset,
                        option,
                        option['values']
                    )

        return queryset



from .fields import ElasticsearchConnectionField
from .types import ElasticsearchObjectType

__title__ = 'graphene-elastic'
__version__ = '0.6.1'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL-2.0-only OR LGPL-2.1-or-later'
__all__ = (
    '__version__',
    'ElasticsearchConnectionField',
    'ElasticsearchObjectType',
)



import graphene
import elasticsearch_dsl

from collections import OrderedDict
from graphene.relay import (
    # Connection,
    Node,
)
from graphene.types.objecttype import ObjectType, ObjectTypeOptions
from graphene.types.utils import yank_fields_from_attrs

# from ..fields import ElasticsearchConnectionField
from ..relay.connection import Connection
from ..converter import convert_elasticsearch_field
from ..registry import Registry, get_global_registry
from ..utils import (
    get_document_fields,
    is_valid_elasticsearch_document,
)

__title__ = 'graphene_elastic.types.elastic_types'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL-2.0-only OR LGPL-2.1-or-later'
__all__ = (
    'construct_fields',
    'construct_self_referenced_fields',
    'ElasticsearchObjectTypeOptions',
    'ElasticsearchObjectType',
)


def construct_fields(document, registry, only_fields, exclude_fields):
    """
    Args:
        document (elasticsearch_dsl.Document):
        registry (graphene_elastic.registry.Registry):
        only_fields ([str]):
        exclude_fields ([str]):

    Returns:
        (OrderedDict, OrderedDict): converted fields and self reference fields.

    """
    _document_fields = get_document_fields(document)
    fields = OrderedDict()
    self_referenced = OrderedDict()
    for name, field in _document_fields.items():
        is_not_in_only = only_fields and name not in only_fields
        is_excluded = name in exclude_fields
        if is_not_in_only or is_excluded:
            # We skip this field if we specify only_fields and is not
            # in there. Or when we exclude this field in exclude_fields
            continue

        # TODO: Finish this once ``List`` is supported (in Elasticsearch
        # it's `multi=True`).
        if field._multi:
            if not (hasattr(field, '_doc_class') or hasattr(field, 'fields')):
                continue

            if hasattr(field, 'fields'):
                for sub_field_name, sub_field in field.fields._d_.items():

                    # Take care of list of self-reference.
                    document_type_obj = sub_field.__class__

                    if document_type_obj == document \
                            or isinstance(document_type_obj, document):
                        self_referenced[name] = field
                        continue
            else:   # if hasattr(field, '_doc_class')
                document_type_obj = field._doc_class

                if document_type_obj == document \
                        or isinstance(document_type_obj, document):
                    self_referenced[name] = field
                    continue

        converted = convert_elasticsearch_field(field, registry)
        if not converted:
            continue
        fields[name] = converted

    return fields, self_referenced


def construct_self_referenced_fields(self_referenced, registry):
    fields = OrderedDict()
    for name, field in self_referenced.items():
        converted = convert_elasticsearch_field(field, registry)
        if not converted:
            continue
        fields[name] = converted

    return fields


def construct_backend_fields(backends, connection):
    """Construct backend fields.

    :param backends:
    :param connection:
    :return:
    """
    backend_fields = OrderedDict()
    for backend_cls in backends:
        backend = backend_cls(connection)
        backend_fields.update(backend.get_backend_document_fields())

    return backend_fields


class ElasticsearchObjectTypeOptions(ObjectTypeOptions):

    document = None  # type: Document
    registry = None  # type: Registry
    connection = None  # type: Type[Connection]
    # filter_backends = []
    # filter_backend_options = {}


class ElasticsearchObjectType(ObjectType):

    @classmethod
    def __init_subclass_with_meta__(cls,
                                    document=None,
                                    registry=None,
                                    skip_registry=False,
                                    only_fields=(),
                                    exclude_fields=(),
                                    connection=None,
                                    connection_class=None,
                                    use_connection=None,
                                    connection_field_class=None,
                                    interfaces=(),
                                    _meta=None,
                                    **options):

        assert is_valid_elasticsearch_document(document), (
            'The attribute document in {}.Meta must be a valid '
            'Elasticsearch-dsl Document. Received "{}" instead.'
        ).format(cls.__name__, type(document))
        if not registry:
            registry = get_global_registry()

        assert isinstance(registry, Registry), (
            'The attribute registry in {}.Meta needs to be an instance of '
            'Registry, received "{}".'
        ).format(cls.__name__, registry)
        converted_fields, self_referenced = construct_fields(
            document, registry, only_fields, exclude_fields
        )
        document_fields = yank_fields_from_attrs(
            converted_fields,
            _as=graphene.Field
        )
        if use_connection is None and interfaces:
            use_connection = any(
                (issubclass(interface, Node) for interface in interfaces)
            )

        if use_connection and not connection:
            # We create the connection automatically
            if not connection_class:
                connection_class = Connection

            connection = connection_class.create_type(
                '{}Connection'.format(cls.__name__), node=cls)

        if connection is not None:
            assert issubclass(connection, Connection), (
                'The attribute connection in {}.Meta must be of type '
                'Connection. Received "{}" instead.'
            ).format(cls.__name__, type(connection))

        if connection_field_class is not None:
            assert issubclass(
                connection_field_class,
                graphene.ConnectionField
            ), (
                'The attribute connection_field_class in {}.Meta must be of '
                'type graphene.ConnectionField. Received "{}" instead.'
            ).format(cls.__name__, type(connection_field_class))
        else:
            from ..fields import ElasticsearchConnectionField
            connection_field_class = ElasticsearchConnectionField

        if _meta:
            assert isinstance(_meta, ElasticsearchObjectTypeOptions), (
                '_meta must be an instance of ElasticsearchObjectTypeOptions, '
                'received {}'
            ).format(_meta.__class__)
        else:
            _meta = ElasticsearchObjectTypeOptions(cls)

        backend_fields = construct_backend_fields(
            backends=options.get('filter_backends', []),
            connection=connection
        )
        document_fields.update(backend_fields)

        _meta.document = document
        _meta.registry = registry
        _meta.fields = document_fields
        _meta.filter_backends = options.get('filter_backends', [])
        _meta.filter_backend_options = dict(options)
        _meta.connection = connection
        _meta.connection_field_class = connection_field_class
        # Save them for later
        _meta.only_fields = only_fields
        _meta.exclude_fields = exclude_fields

        super(ElasticsearchObjectType, cls).__init_subclass_with_meta__(
            _meta=_meta, interfaces=interfaces, **options
        )
        if not skip_registry:
            registry.register(cls)
            # Notes: Take care list of self-reference fields.
            converted_fields = construct_self_referenced_fields(
                self_referenced,
                registry
            )
            if converted_fields:
                document_fields = yank_fields_from_attrs(
                    converted_fields,
                    _as=graphene.Field
                )
                cls._meta.fields.update(document_fields)
                registry.register(cls)

    @classmethod
    def rescan_fields(cls):
        """Attempts to rescan fields and will insert any not converted
        initially."""

        converted_fields, self_referenced = construct_fields(
            cls._meta.document, cls._meta.registry,
            cls._meta.only_fields, cls._meta.exclude_fields
        )

        document_fields = yank_fields_from_attrs(
            converted_fields,
            _as=graphene.Field
        )

        # The initial scan should take precedence
        for field in document_fields:
            if field not in cls._meta.fields:
                cls._meta.fields.update({field: document_fields[field]})
        # Self-referenced fields can't change between scans!

    @classmethod
    def is_type_of(cls, root, info):
        if isinstance(root, cls):
            return True

        if not is_valid_elasticsearch_document(type(root)):
            raise Exception((
                'Received incompatible instance "{}".'
            ).format(root))
        return isinstance(root, cls._meta.document)

    @classmethod
    def get_node(cls, info, id):
        return cls._meta.document.get(id)

    @property
    def id(self):
        return self.meta.id

    def resolve_id(self, info):
        return self.meta.id



# """
# Transitional compatibility module. Contains various field wrappers and
# helpers for painless (testing of) Elastic 6.x to Elastic 7.x transition. This
# module is not supposed to solve all transition issues for you. Better move to
# Elastic 7.x as soon as possible.
# """
# __title__ = 'graphene_elastic.compat'
# __author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
# __copyright__ = '2019-2020 Artur Barseghyan'
# __license__ = 'GPL 2.0/LGPL 2.1'
# __all__ = (
#     'nested_sort_entry',
# )
#
#
# def nested_sort_entry(path):
#     """String field.
#         :param path: Full path to nested container, separated by period
#         :type: str
#         :return: Dictionary of full nested path
#         :rtype: dict
#         """
#     from .versions import get_elasticsearch_version
#     version = get_elasticsearch_version()
#     if version[0] < 6 or (version[0] == 6 and version[1] < 1):
#         return {'nested_path': path}
#     nested_path = {}
#     path_list = path.split('.')
#     for _ in reversed(path_list):
#         if nested_path:
#             nested_path = {'path': '.'.join(path_list), 'nested': nested_path}
#         else:
#             nested_path = {'path': '.'.join(path_list)}
#         path_list.pop()
#     return {'nested': nested_path}



import graphene
from ..enums import NoValue

__title__ = "graphene_elastic.filter_backends.queries"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    'Contains',
    'Direction',
    'EndsWith',
    'Exclude',
    'Exists',
    'GeoBoundingBox',
    'GeoDistance',
    'GeoPolygon',
    'Gt',
    'Gte',
    'In',
    'IsNull',
    'Lt',
    'Lte',
    'Point',
    'Prefix',
    'Range',
    'StartsWith',
    'Term',
    'Terms',
    'Wildcard',
)


class _ListOfTypeString(graphene.List):
    """List of type string."""

    required = True

    def __init__(self, *args, **kwargs):
        super(_ListOfTypeString, self).__init__(
            of_type=graphene.String,
            *args,
            **kwargs
        )


_number_or_date_types = (
    graphene.Decimal,
    graphene.Float,
    graphene.Int,
    graphene.Date,
    graphene.DateTime,
)


class _NumberOrDate(graphene.InputObjectType):
    """Number or Date."""

    decimal = graphene.Decimal(required=False)
    float = graphene.Float(required=False)
    int = graphene.Int(required=False)
    datetime = graphene.DateTime(required=False)
    date = graphene.Date(required=False)

# ***************************************************************
# ************************* Filtering ***************************
# ***************************************************************


class Term(graphene.String):
    """Terms.

        filter:[
            {category: {term: "Python"}]}}
        ]
    """

    required = True


class Terms(_ListOfTypeString):
    """Terms.

        filter:[
            {category: {terms: ["Python", "Django"]}}
        ]
    """


class Range(graphene.InputObjectType):
    """Range.

        filter:[
            {range: {lower: "1000", upper: "2000", boost: "2.0"}]}}
        ]
    """
    lower = graphene.Field(_NumberOrDate, required=True)
    upper = graphene.Field(_NumberOrDate)
    boost = graphene.Decimal()


class Exists(graphene.Boolean):
    """Exists.

        filter:[
            {category: {exist: true}]}}
        ]
    """

    required = True


class Prefix(graphene.String):
    """Prefix.

        filter:[
            {category: {prefix: "bio"}]}}
        ]
    """

    required = True


class StartsWith(Prefix):
    """Starts with (alias of prefix)."""


class Wildcard(graphene.String):
    """Wildcard.

        filter:[
            {category: {wildcard: "*bio*"}]}}
        ]
    """

    required = True


class GeoDistance(graphene.InputObjectType):
    """Geo distance.

        filter:[
            {place: {geoDistance: {distance: "9km", lat: "40.0", lon="70.0"}}}
        ]
    """

    distance = graphene.String(required=True)
    lat = graphene.Decimal(required=True)
    lon = graphene.Decimal(required=True)


class Point(graphene.InputObjectType):
    """Point. Helper for many geo lookups."""

    lat = graphene.Decimal(required=True)
    lon = graphene.Decimal(required=True)


class GeoPolygon(graphene.InputObjectType):
    """Geo polygon.

        filter:[
            {place: {geoPolygon: {points: [{lat: "40.0", lon="70.0"}]}}}
        ]
    """

    points = graphene.List(Point, required=True)


class GeoBoundingBox(graphene.InputObjectType):
    """Geo polygon.

        filter:[
            {place: {geoBoundingBox: {
                topLeft: {lat: "40.0", lon="70.0"},
                bottomRight: {lat: "80.0", lon: "90.0"}
            }}}
        ]
    """

    top_left = graphene.Field(Point, required=True)
    bottom_right = graphene.Field(Point, required=True)


class Contains(graphene.String):
    """Wildcard.

        filter:[
            {category: {contains: "lish"}]}}
        ]
    """

    required = True


class In(_ListOfTypeString):
    """In.

        filter:[
            {category: {in: ["Python", "Django"]}}
        ]
    """


class Gt(_NumberOrDate):
    """Gt.

        filter:[
            {category: {gt: "1"}]}}
        ]
    """

    class Meta:
        types = list(_number_or_date_types)

    required = True


class Gte(_NumberOrDate):
    """Gte.

        filter:[
            {category: {gte: "1"}]}}
        ]
    """

    class Meta:
        types = list(_number_or_date_types)

    required = True


class Lt(_NumberOrDate):
    """Lt.

        filter:[
            {category: {lt: "1"}]}}
        ]
    """

    class Meta:
        types = list(_number_or_date_types)

    required = True


class Lte(_NumberOrDate):
    """Lte.

        filter:[
            {category: {lte: "1"}]}}
        ]
    """

    class Meta:
        types = list(_number_or_date_types)

    required = True


class EndsWith(graphene.String):
    """Ends with.

        filter:[
            {category: {endsWith: "dren"}]}}
        ]
    """

    required = True


class IsNull(graphene.Boolean):
    """Is null.

        filter:[
            {category: {isNull: true}]}}
        ]
    """

    required = True


class Exclude(_ListOfTypeString):
    """Exclude.

        filter:[
            {category: {exclude: ["Python", "Django"]}}
        ]
    """

# ***************************************************************
# ************************** Ordering ***************************
# ***************************************************************


@graphene.Enum.from_enum
class Direction(NoValue):
    """Ordering direction."""

    ASC = 'asc'
    DESC = 'desc'



from setuptools import find_packages, setup

setup(
    name="graphene-elastic",
    version="0.6.1",
    description="Graphene Elasticsearch (DSL) integration",
    long_description=open("README.rst").read(),
    url="https://github.com/barseghyanartur/graphene-elastic",
    project_urls={
        "Bug Tracker": "https://github.com/barseghyanartur/graphene-elastic/",
        "Documentation": "https://graphene-elastic.readthedocs.io/",
        "Source Code": "https://github.com/barseghyanartur/graphene-elastic/",
        "Changelog": "https://graphene-elastic.readthedocs.io/"
                     "en/latest/changelog.html",
    },
    author="Artur Barseghyan",
    author_email="artur.barseghyan@gmail.com",
    license="GPL 2.0/LGPL 2.1",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="api graphql protocol rest relay graphene elasticsearch "
    "elasticsearch-dsl",
    package_dir={"": "src"},
    packages=find_packages(
        where="./src",
        exclude=["tests"]
    ),
    install_requires=[
        "graphene>=2.1.3,<3",
        "elasticsearch>=6.0",
        "elasticsearch-dsl>=6.0",
        "singledispatch>=3.4.0.3",
        "iso8601>=0.1.12",
        "stringcase",
    ],
    python_requires=">=3.6",
    zip_safe=True,
    tests_require=[
        "pytest>=3.3.2",
        "mock",
    ],
    include_package_data=True,
)



from __future__ import absolute_import

# from decimal import Decimal

from collections import Iterable, OrderedDict
from functools import partial, reduce

import graphene
# from graphene import NonNull
# from graphql_relay import connection_from_list
import elasticsearch_dsl
from promise import Promise
from graphene.relay import ConnectionField, PageInfo
from graphene.types.argument import to_arguments
from graphene.types.dynamic import Dynamic
from graphene.types.structures import Structure

from .advanced_types import (
    FileFieldType,
    PointFieldType,
    MultiPolygonFieldType,
)
from .arrayconnection import connection_from_list_slice
from .converter import (
    convert_elasticsearch_field,
    ElasticsearchConversionError,
)
# from .filter_backends import (
#     SearchFilterBackend,
#     FilteringFilterBackend,
#     OrderingFilterBackend,
#     DefaultOrderingFilterBackend,
# )
from .logging import logger
from .registry import get_global_registry
from .settings import graphene_settings
from .types import ElasticsearchObjectType
from .utils import get_node_from_global_id  # get_model_reference_fields

__title__ = "graphene_elastic.fields"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("ElasticsearchConnectionField",)


# def json_default(obj):
#     if isinstance(obj, Decimal):
#         return str(obj)  # String version
#     return obj


class ElasticsearchConnectionField(ConnectionField):
    def __init__(self, type, *args, **kwargs):
        self.on = kwargs.pop("on", False)  # From graphene-django
        self.max_limit = kwargs.pop(
            "max_limit", graphene_settings.RELAY_CONNECTION_MAX_LIMIT
        )  # From graphene-django
        self.enforce_first_or_last = kwargs.pop(
            "enforce_first_or_last",
            graphene_settings.RELAY_CONNECTION_ENFORCE_FIRST_OR_LAST,
        )  # From graphene-django
        get_queryset = kwargs.pop("get_queryset", None)
        if get_queryset:
            assert callable(
                get_queryset
            ), "Attribute `get_queryset` on {} must be callable.".format(self)
        self._get_queryset = get_queryset

        super(ElasticsearchConnectionField, self).__init__(
            type, *args, **kwargs
        )

    @property
    def type(self):
        # from .types import ElasticsearchObjectType
        _type = super(ConnectionField, self).type
        assert issubclass(
            _type, ElasticsearchObjectType
        ), "ElasticsearchConnectionField only accepts " \
           "ElasticsearchObjectType types"
        assert (
            _type._meta.connection
        ), "The type {} doesn't have a connection".format(_type.__name__)
        return _type._meta.connection

    # @property
    # def connection_type(self):  # From graphene-django
    #     type = self.type
    #     if isinstance(type, NonNull):
    #         return type.of_type
    #     return type

    @property
    def node_type(self):
        return self.type._meta.node

    @property
    def document(self):
        return self.node_type._meta.document

    @property
    def doc_type(self):
        return self.document._doc_type

    # def get_manager(self):  # From graphene-django
    #     if self.on:
    #         return getattr(self.document, self.on)
    #     else:
    #         return self.document.search()

    @property
    def registry(self):
        return getattr(self.node_type._meta, "registry", get_global_registry())

    @property
    def args(self):
        return to_arguments(
            self._base_args or OrderedDict(),
            dict(self.field_args, **self.reference_args),
        )

    @property
    def default_filter_backends(self):
        return [
            # SearchFilterBackend,
            # FilteringFilterBackend,
            # OrderingFilterBackend,
            # DefaultOrderingFilterBackend,
        ]

    @property
    def filter_backends(self):
        return getattr(
            self.node_type._meta,
            "filter_backends",
            self.default_filter_backends,
        )

    @args.setter
    def args(self, args):
        self._base_args = args

    def _field_args(self, items):
        def is_filterable(k):
            """
            Args:
                k (str): field name.
            Returns:
                bool
            """
            if k not in self.doc_type.mapping.properties.properties._d_:
                return False
            try:
                converted = convert_elasticsearch_field(
                    self.doc_type.mapping.properties.properties._d_.get(k),
                    self.registry,
                )
            except ElasticsearchConversionError:
                return False
            if isinstance(converted, (ConnectionField, Dynamic)):
                return False
            if callable(getattr(converted, "type", None)) and isinstance(
                converted.type(),
                (
                    FileFieldType,
                    PointFieldType,
                    MultiPolygonFieldType,
                    graphene.Union,
                ),
            ):
                return False
            return True

        def get_type(v):
            if isinstance(v.type, Structure):
                return v.type.of_type()
            return v.type()

        # Filter fields are here: self.node_type._meta.filter_fields
        # Search fields are here: self.node_type._meta.search_fields

        params = {}

        for backend_cls in self.filter_backends:
            if backend_cls.has_query_fields:
                backend = backend_cls(self)
                _query_fields = backend.get_backend_query_fields(
                    items=items,
                    is_filterable_func=is_filterable,
                    get_type_func=get_type,
                )
                if _query_fields:
                    params.update(_query_fields)

        return params

    @property
    def field_args(self):
        return self._field_args(list(self.fields.items()))

    @property
    def reference_args(self):
        def get_reference_field(r, kv):
            field = kv[1]
            # TODO: Find out whether this is applicable to Elasticsearch
            if callable(getattr(field, "get_type", None)):
                _type = field.get_type()
                if _type:
                    node = _type._type._meta
                    if "id" in node.fields and not issubclass(
                        node.document, (elasticsearch_dsl.InnerDoc,)
                    ):
                        r.update({kv[0]: node.fields["id"]._type.of_type()})
            return r

        return reduce(get_reference_field, self.fields.items(), {})

    @property
    def fields(self):
        # We might need self._type._doc_type.mapping.properties.properties._d_
        return self._type._meta.fields

    def get_queryset(self, document, info, **args):
        if args:
            # reference_fields = get_model_reference_fields(self.model)
            reference_fields = {}
            hydrated_references = {}
            for arg_name, arg in args.copy().items():
                if arg_name in reference_fields:
                    reference_obj = get_node_from_global_id(
                        reference_fields[arg_name], info, args.pop(arg_name)
                    )
                    hydrated_references[arg_name] = reference_obj
            args.update(hydrated_references)

        if self._get_queryset:
            queryset_or_filters = self._get_queryset(document, info, **args)
            if isinstance(queryset_or_filters, elasticsearch_dsl.Search):
                return queryset_or_filters
            else:
                args.update(queryset_or_filters)
        qs = document.search()

        for backend_cls in self.filter_backends:
            backend = backend_cls(self, args=dict(args))
            qs = backend.filter(qs)

        try:
            logger.debug(qs.to_dict())
        except Exception as err:
            logger.debug(err)
        return qs

    def default_resolver(self, _root, info, **args):
        args = args or {}
        connection_args = {
            "first": args.pop("first", None),
            "last": args.pop("last", None),
            "before": args.pop("before", None),
            "after": args.pop("after", None),
            "max_limit": args.pop(
                "max_limit",
                graphene_settings.RELAY_CONNECTION_MAX_LIMIT
            ),
            "enforce_first_or_last": args.pop(
                "enforce_first_or_last",
                graphene_settings.RELAY_CONNECTION_ENFORCE_FIRST_OR_LAST
            ),
        }

        _id = args.pop("id", None)

        if _id is not None:
            iterables = [get_node_from_global_id(self.node_type, info, _id)]
            list_length = 1
        # TODO: The next line never happens. We might want to make sure
        # functionality that must be there is present
        elif callable(getattr(self.document, "search", None)):
            iterables = self.get_queryset(self.document, info, **args)
            list_length = iterables.count()
        else:
            iterables = []
            list_length = 0
        connection = connection_from_list_slice(
            list_slice=iterables,
            args=connection_args,
            list_length=list_length,
            list_slice_length=list_length,
            connection_type=self.type,
            edge_type=self.type.Edge,
            pageinfo_type=graphene.PageInfo,
            connection_field=self
        )
        connection.iterable = iterables
        connection.list_length = list_length
        return connection

    def chained_resolver(self, resolver, is_partial, root, info, **args):
        if not bool(args) or not is_partial:
            resolved = resolver(root, info, **args)
            if resolved is not None:
                return resolved
        return self.default_resolver(root, info, **args)

    @classmethod
    def resolve_connection(cls,
                           connection_type,
                           args,
                           resolved,
                           connection_field=None):
        if isinstance(resolved, connection_type):
            return resolved

        assert isinstance(resolved, Iterable), (
            "Resolved value from the connection field have to be iterable or "
            "instance of {}. "
            'Received "{}"'
        ).format(connection_type, resolved)

        _len = resolved.hits.total["value"]

        connection = connection_from_list_slice(
            resolved.hits,
            args,
            slice_start=0,
            list_length=_len,
            list_slice_length=_len,
            connection_type=connection_type,
            edge_type=connection_type.Edge,
            pageinfo_type=PageInfo,
            connection_field=connection_field
        )
        connection.iterable = resolved
        connection.length = _len
        return connection

    @classmethod
    def connection_resolver(cls,
                            resolver,
                            connection_type,
                            root,
                            info,
                            connection_field=None,
                            **args):
        first = args.get("first")
        last = args.get("last")
        enforce_first_or_last = args.get("enforce_first_or_last")
        max_limit = args.get("max_limit")
        # connection_field = args.get("connection_field")

        if enforce_first_or_last:
            assert first or last, (
                "You must provide a `first` or `last` value to properly "
                "paginate the `{}` connection."
            ).format(info.field_name)

        if max_limit:
            if first:
                assert first <= max_limit, (
                    "Requesting {} records on the `{}` connection exceeds "
                    "the `first` limit of {} records."
                ).format(first, info.field_name, max_limit)
                args["first"] = min(first, max_limit)

            if last:
                assert last <= max_limit, (
                    "Requesting {} records on the `{}` connection exceeds "
                    "the `last` limit of {} records."
                ).format(last, info.field_name, max_limit)
                args["last"] = min(last, max_limit)

        iterable = resolver(root, info, **args)
        if isinstance(connection_type, graphene.NonNull):
            connection_type = connection_type.of_type

        on_resolve = partial(
            cls.resolve_connection,
            connection_type,
            args,
            connection_field=connection_field
        )

        if Promise.is_thenable(iterable):
            return Promise.resolve(iterable).then(on_resolve)

        return on_resolve(iterable)

    def get_resolver(self, parent_resolver):
        super_resolver = self.resolver or parent_resolver
        resolver = partial(
            self.chained_resolver,
            super_resolver,
            isinstance(super_resolver, partial),
        )
        return partial(
            self.connection_resolver,
            resolver,
            self.type,
            max_limit=self.max_limit,
            enforce_first_or_last=self.enforce_first_or_last,
            connection_field=self
        )



import time
import unittest
import factories
from .base import BaseGrapheneElasticTestCase

__all__ = (
    'FacetedSearchBackendElasticTestCase',
)


class FacetedSearchBackendElasticTestCase(BaseGrapheneElasticTestCase):

    def setUp(self):
        super(FacetedSearchBackendElasticTestCase, self).setUp()
        self.alice = "Alice"
        self.alice_category = 'Delusional'
        self.num_alice_posts = 9
        self.alice_posts = factories.PostFactory.create_batch(
            self.num_alice_posts,
            category=self.alice_category

        )
        for _post in self.alice_posts:
            _post.content = "{} {} {}".format(
                self.faker.paragraph(),
                self.alice,
                self.faker.paragraph()
            )
            _post.save()

        self.beast = "Jabberwocky"
        self.beast_category = 'Insanity'
        self.num_beast_posts = 5
        self.beast_posts = factories.PostFactory.create_batch(
            self.num_beast_posts,
            category=self.beast_category,
        )
        for _post in self.beast_posts:
            _post.content = "{} {} {}".format(
                self.faker.paragraph(),
                self.beast,
                self.faker.paragraph()
            )
            _post.save()

        self.num_other_posts = 40
        self.other_posts = factories.PostFactory.create_batch(
            self.num_other_posts
        )
        for _post in self.other_posts:
            _post.save()

        time.sleep(2)

    def __test_faceted_search_no_facets_args(self):
        """Test faceted search.

        :param num_posts:
        :return:
        """
        query = """
        query {
          allPostDocuments {
            facets
            edges {
              node {
                category
                title
                comments
              }
            }
          }
        }
        """
        print(query)
        executed = self.client.execute(query)
        data = executed.get('data', {}).get('allPostDocuments', {})
        self.assertIn('facets', data)
        self.assertIn('tags', data['facets'])
        self.assertNotIn('category', data['facets'])
        self.assertIn('aggs', data['facets']['tags'])
        self.assertIn('buckets', data['facets']['tags']['aggs'])

    def __test_faceted_search_with_facets_args(self):
        """Test faceted search.

        :return:
        """
        query = """
        query {
          allPostDocuments(facets:[category]) {
            facets
            edges {
              node {
                category
                title
                comments
              }
            }
          }
        }
        """
        print(query)
        executed = self.client.execute(query)
        data = executed.get('data', {}).get('allPostDocuments', {})
        self.assertIn('facets', data)
        self.assertIn('tags', data['facets'])
        self.assertIn('category', data['facets'])
        self.assertIn('aggs', data['facets']['category'])
        self.assertIn('aggs', data['facets']['tags'])
        self.assertIn('buckets', data['facets']['category']['aggs'])
        self.assertIn('buckets', data['facets']['tags']['aggs'])

        buckets = data['facets']['category']['aggs']['buckets']
        doc_counts = {_d['key']: _d['doc_count'] for _d in buckets}
        self.assertEqual(
            doc_counts.get(self.alice_category),
            self.num_alice_posts
        )
        self.assertEqual(
            doc_counts.get(self.beast_category),
            self.num_beast_posts
        )

    def _test_faceted_search(self):
        """"Test search content.

        :return:
        """
        # Covering default facets (no specific facets enabled)
        with self.subTest('Test empty facets'):
            self.__test_faceted_search_no_facets_args()

        # Covering specific facets (that aren't enabled by default)
        with self.subTest('Test given facets'):
            self.__test_faceted_search_with_facets_args()

    def test_all(self):
        """Test all.

        Since we don't write in specific tests, it's more efficient to run
        them all from a single method in order to save on speed ups between
        tests.
        """
        self._test_faceted_search()


if __name__ == '__main__':
    unittest.main()



from copy import deepcopy
# from collections import OrderedDict
import enum

from elasticsearch_dsl import TermsFacet
from elasticsearch_dsl.query import Q

import graphene
from stringcase import pascalcase as to_pascal_case
# from graphene.types import Field

from ...constants import DYNAMIC_CLASS_NAME_PREFIX
from ..base import BaseBackend

__title__ = 'graphene_elastic.filter_backends.faceted_search.common'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019-2020 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'FacetedSearchFilterBackend',
)


def default_agg_field_name_getter(field_name):
    return '_filter_' + field_name


def default_agg_bucket_name_getter(field_name):
    return field_name


class FacetedSearchFilterBackend(BaseBackend):
    """Faceted search filter backend."""

    prefix = 'facets'
    has_connection_fields = True
    has_query_fields = True

    # def get_backend_connection_fields(self):
    #     """Get backend connection fields.
    #
    #     Typical use-case - a backend that alters the Connection object
    #     and adds additional fields next to `edges` and `pageInfo` (see
    #     the `graphene_elastic.relay.connection.Connection` for more
    #     information).
    #
    #     :rtype dict:
    #     :return:
    #     """
    #     from ...types.json_string import JSONString
    #     return OrderedDict([
    #         (
    #             "facets",
    #             Field(
    #                 JSONString,
    #                 name="facets",
    #                 required=False,
    #                 description="Pagination data for this connection.",
    #             ),
    #         ),
    #     ])

    def alter_connection(self, connection, slice):
        """Alter connection object.

        You can add various properties here, returning the altered object.
        Typical use-case would be adding facets to the connection.

        :param connection:
        :param slice:
        :return:
        """
        try:
            connection.facets = slice.aggregations
        except Exception as err:
            connection.facets = {}

    @property
    def faceted_search_fields(self):
        """Faceted search filter fields."""
        search_fields = getattr(
            self.connection_field.type._meta.node._meta,
            'filter_backend_options',
            {}
        ).get('faceted_search_fields', {})
        return deepcopy(search_fields)

    def field_belongs_to(self, field_name):
        """Check if given filter field belongs to the backend.

        :param field_name:
        :return:
        """
        return field_name in self.faceted_search_fields

    def get_backend_query_fields(self,
                                 items,
                                 is_filterable_func,
                                 get_type_func):
        """Construct backend filtering fields.

        :param items:
        :param is_filterable_func:
        :param get_type_func:
        :return:
        """
        params = {}
        for field, value in items:
            if is_filterable_func(field):
                # Getting other backend specific fields (schema dependant)
                if self.field_belongs_to(field):
                    params.update({field: field})
        return {
            self.prefix: graphene.Argument(
                graphene.List(
                    graphene.Enum.from_enum(
                        enum.Enum(
                            "{}{}{}BackendEnum".format(
                                DYNAMIC_CLASS_NAME_PREFIX,
                                to_pascal_case(self.prefix),
                                self.connection_field.type.__name__
                            ),
                            params
                        )
                    )
                )
            )
        }

    def prepare_faceted_search_fields(self):
        """Prepare faceted search fields.

        Prepares the following structure:

            >>> {
            >>>     'publisher': {
            >>>         'field': 'publisher.raw',
            >>>         'facet': TermsFacet,
            >>>         'enabled': False,
            >>>     }
            >>>     'date_published': {
            >>>         'field': 'date_published.raw',
            >>>         'facet': DateHistogramFacet,
            >>>         'options': {
            >>>             'interval': 'month',
            >>>         },
            >>>         'enabled': True,
            >>>     },
            >>> }

        :return: Faceted search fields options.
        :rtype: dict
        """
        if not self.faceted_search_fields:
            return {}

        faceted_search_args = dict(self.args).get(self.prefix, [])

        faceted_search_fields = deepcopy(self.faceted_search_fields)

        for field, options in faceted_search_fields.items():
            if options is None or isinstance(options, str):
                faceted_search_fields[field] = {
                    'field': options or field
                }
            elif 'field' not in faceted_search_fields[field]:
                faceted_search_fields[field]['field'] = field

            if 'enabled' in faceted_search_fields[field] \
                    or field in faceted_search_args:
                faceted_search_fields[field]['enabled'] = True
            else:
                faceted_search_fields[field]['enabled'] = False

            if 'facet' not in faceted_search_fields[field]:
                faceted_search_fields[field]['facet'] = TermsFacet

            if 'options' not in faceted_search_fields[field]:
                faceted_search_fields[field]['options'] = {}

            faceted_search_fields[field]['global'] = \
                faceted_search_fields[field].get('global', False)

        return faceted_search_fields

    def get_faceted_search_query_params(self):
        """Get highlight query params.

        :return: List of search query params.
        :rtype: list
        """
        faceted_search_args = dict(self.args).get(self.prefix, {})
        return faceted_search_args

    def construct_facets(self):
        """Construct facets.

        Turns the following structure:

            >>> {
            >>>     'publisher': {
            >>>         'field': 'publisher.raw',
            >>>         'facet': TermsFacet,
            >>>         'enabled': False,
            >>>     }
            >>>     'date_published': {
            >>>         'field': 'date_published',
            >>>         'facet': DateHistogramFacet,
            >>>         'options': {
            >>>             'interval': 'month',
            >>>         },
            >>>         'enabled': True,
            >>>     },
            >>> }

        Into the following structure:

            >>> {
            >>>     'publisher': TermsFacet(field='publisher.raw'),
            >>>     'publishing_frequency': DateHistogramFacet(
            >>>         field='date_published.raw',
            >>>         interval='month'
            >>>     ),
            >>> }
        """
        _facets = {}
        faceted_search_query_params = self.get_faceted_search_query_params()
        faceted_search_fields = self.prepare_faceted_search_fields()
        for _field, _options in faceted_search_fields.items():
            if _field in faceted_search_query_params or _options['enabled']:
                _facets.update(
                    {
                        _field: {
                            'facet': faceted_search_fields[_field]['facet'](
                                field=faceted_search_fields[_field]['field'],
                                **faceted_search_fields[_field]['options']
                            ),
                            'global': faceted_search_fields[_field]['global'],
                        }
                    }
                )
        return _facets

    def aggregate(self,
                  queryset,
                  agg_field_name_getter=default_agg_field_name_getter,
                  agg_bucket_name_getter=default_agg_bucket_name_getter):
        """Aggregate.

        :param queryset:
        :param agg_field_name_getter: callable.
        :param agg_bucket_name_getter:
        :return:
        """
        _facets = self.construct_facets()
        for _field, _facet in _facets.items():
            agg = _facet['facet'].get_aggregation()
            agg_filter = Q('match_all')

            # TODO: Implement
            # for __filter_field, __filter in self._filters.items():
            #     if __field == __filter_field:
            #         continue
            #     agg_filter &= __filter

            if _facet['global']:
                queryset.aggs.bucket(
                    agg_field_name_getter(_field),
                    'global'
                ).bucket(
                    agg_bucket_name_getter(_field),
                    agg
                )
            else:
                queryset.aggs.bucket(
                    agg_field_name_getter(_field),
                    'filter',
                    filter=agg_filter
                ).bucket(
                    agg_bucket_name_getter(_field),
                    agg
                )

        return queryset

    def filter(self, queryset):
        """Filter the queryset.

        :param queryset: Base queryset.
        :type queryset: elasticsearch_dsl.search.Search
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        return self.aggregate(
            queryset,
            agg_field_name_getter=lambda field: field,
            agg_bucket_name_getter=lambda field: 'aggs'
        )



from collections import OrderedDict
import graphene
from stringcase import pascalcase as to_pascal_case

from ..constants import (
    DYNAMIC_CLASS_NAME_PREFIX,
    SEPARATOR_LOOKUP_NAME,
    SEPARATOR_LOOKUP_FILTER,
    SEPARATOR_LOOKUP_COMPLEX_VALUE,
    SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE,
)

__title__ = "graphene_elastic.filter_backends.base"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019-2020 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("BaseBackend",)


class BaseBackend(object):
    """Base backend."""

    # Prefix for the backend. This is used to isolate the backend variables
    # but also is used in dynamically constructing the class names of
    # `graphene` types.
    prefix = None

    # If set to True, an attempt will be made to construct the GraphQL query
    # fields for backend.
    has_query_fields = False

    # If set to True, an attempt will be made to construct the connection
    # fields for backend.
    has_connection_fields = False

    def __init__(self, connection_field, args=None):
        self.connection_field = connection_field
        self.args = args or {}
        assert self.prefix

    def field_belongs_to(self, field_name):
        """Check if given filter field belongs to the backend.

        :param field_name:
        :return:
        """
        raise NotImplementedError

    # def add_arg_prefix(self, arg_name):
    #     return "{}_{}".format(self.prefix, arg_name)
    #
    # def get_field_name(self, arg_name):
    #     """Get field name.
    #
    #     :param arg_name:
    #     :return:
    #     """
    #     if arg_name.startswith(self.prefix):
    #         return arg_name.lstrip("{}_".format(self.prefix))
    #
    # def arg_belongs_to(self, arg_name):
    #     field_name = self.get_field_name(arg_name)
    #     if field_name:
    #         return self.field_belongs_to(field_name)
    #     return False

    def filter(self, queryset):
        """Filter. This method alters current queryset.

        :param queryset:
        :return:
        """
        raise NotImplementedError

    @property
    def doc_type(self):
        """Shortcut to the Elasticsearch document type.

        :return:
        """
        return self.connection_field.document._doc_type

    @classmethod
    def generic_query_fields(cls):
        """Generic backend specific query fields.

        For instance, for search filter backend it would be
        ``{'search': String()}``.

        :return:
        :rtype dict:
        """
        return {}

    def get_backend_document_fields(self):
        """Get additional document fields for the backend.

        For instance, the ``Highlight`` backend add additional field named
        ``highlight`` to the list of fields.

        Sample query:

            query {
              allPostDocuments(search:{title:{value:"alice"}}) {
                edges {
                  node {
                    id
                    title
                    highlight
                  }
                }
              }
            }

        Sample response:

            {
              "data": {
                "allPostDocuments": {
                  "edges": [
                    {
                      "node": {
                        "id": "UG9zdDp5a1ppVlcwQklwZ2dXbVlJQV91Rw==",
                        "title": "thus Alice style",
                        "highlight": {
                          "title": [
                            "thus <b>Alice</b> style"
                          ]
                        }
                      }
                    },
                    ...
                  ]
                }
              }
            }

        That ``highlight`` part on both sample query and sample response
        isn't initially available on the connection level, but added with
        help of the filter backend.
        :return:
        """
        return OrderedDict()

    def get_field_type(self, field_name, field_value, base_field_type):
        """Get field type.

        :return:
        """
        raise NotImplementedError

    def get_backend_default_query_fields_params(self):
        """Get default query fields params for the backend.

        :rtype: dict
        :return:
        """
        return {}

    def get_backend_query_fields(self,
                                 items,
                                 is_filterable_func,
                                 get_type_func):
        """Construct backend query fields.

        :param items:
        :param is_filterable_func:
        :param get_type_func:
        :return:
        """
        params = self.get_backend_default_query_fields_params()
        for field, value in items:
            if is_filterable_func(field):
                # Getting other backend specific fields (schema dependant)
                if self.field_belongs_to(field):
                    params.update({
                        field: self.get_field_type(
                            field,
                            value,
                            get_type_func(value)
                        )
                    })

        return {
            self.prefix: graphene.Argument(
                type(
                    "{}{}{}BackendFilter".format(
                        DYNAMIC_CLASS_NAME_PREFIX,
                        to_pascal_case(self.prefix),
                        self.connection_field.type.__name__
                    ),
                    (graphene.InputObjectType,),
                    params,
                )
            )
        }

    def get_backend_connection_fields_type(self):
        """Get backend connection fields type.

        Typical use-case - a backend that alters the Connection object
        and adds additional fields next to `edges` and `pageInfo` (see
        the `graphene_elastic.relay.connection.Connection` for more
        information).

        :return:
        """
        return {}

    def get_backend_connection_fields(self):
        """Get backend connection fields.

        Typical use-case - a backend that alters the Connection object
        and adds additional fields next to `edges` and `pageInfo` (see
        the `graphene_elastic.relay.connection.Connection` for more
        information).

        :rtype dict:
        :return:
        """
        return {}

    def alter_connection(self, connection, slice):
        """Alter connection object.

        You can add various properties here, returning the altered object.
        Typical use-case would be adding facets to the connection.

        :param connection:
        :param slice:
        :return:
        """
        return connection

    @classmethod
    def split_lookup_name(cls, value, maxsplit=-1):
        """Split lookup value.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup value split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_NAME, maxsplit)

    @classmethod
    def split_lookup_filter(cls, value, maxsplit=-1):
        """Split lookup filter.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup filter split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_FILTER, maxsplit)

    @classmethod
    def split_lookup_complex_value(cls, value, maxsplit=-1):
        """Split lookup complex value.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup filter split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_COMPLEX_VALUE, maxsplit)

    @classmethod
    def split_lookup_complex_multiple_value(cls, value, maxsplit=-1):
        """Split lookup complex multiple value.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup filter split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE, maxsplit)

    @classmethod
    def apply_filter(cls, queryset, options=None, args=None, kwargs=None):
        """Apply filter.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return queryset.filter(*args, **kwargs)

    @classmethod
    def apply_query(cls, queryset, options=None, args=None, kwargs=None):
        """Apply query.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return queryset.query(*args, **kwargs)
